# these are wrappers for  read_ncdf_output_files equivalents but load from runcase_info file name
import numpy as np

from oceantracker.util import json_util
from read_oceantracker.python import read_ncdf_output_files
from os import path
from glob import glob

def get_case_info_file_from_run_file(runInfo_fileName_or_runInfoDict, ncase = 0,
                                     run_output_dir= None):
    # get case_info.json file name from runInfo dict or json file name, can also set root_output_dir, if output moved to new location
    # ncase is one based
    if type(runInfo_fileName_or_runInfoDict) is str:
       run_case_info = json_util.read_JSON(runInfo_fileName_or_runInfoDict)
    else:
        run_case_info = runInfo_fileName_or_runInfoDict

    if run_case_info is None:
        # file not found
       raise IOError('Could not read file' +runInfo_fileName_or_runInfoDict)

    if run_output_dir is None:
        run_output_dir = run_case_info['output_files']['run_output_dir']

    try:
        case_info_file_name = path.join(run_output_dir, run_case_info)

    except Exception as e:
        raise Exception('load_ouput_files.get_case_info: Can not find case file json "' + str(case_info_file_name) + "', file missing or case may have had an error see *.err file")

    return path.normpath(case_info_file_name)

def get_case_info_files_from_dir(dir_name, case = None):
    # get all case_info files in given folder, optionally only return given case file (first case is case = 1)
    # folder is cleaned out before start of run so any unfinished caseInfo.json files will  be missing
    # this method allows for missing cases from incomplete runs done in parallel, so  items in list = None for missing cases

    mask= path.join(dir_name, '*_caseInfo.json')
    case_file_names= glob(mask)

    if len(case_file_names)==0:  raise IOError('No case info files in dir '  + dir_name + ', matching mask *_caseInfo.json')

    # get cases as process IDs matching mask
    case_file_list, caseID = [], []
    for case_name in case_file_names:
        case_file_list.append(case_name)
        d = read_case_info_file(case_name)
        caseID.append(d['caseID'])

    # make a full list of cases from their caseID's,  which allows for any missing cases due to dead rund
    case_files_out = max(caseID+1)*[None]
    for ID, c in zip(caseID, case_file_list):
        case_files_out[ID] = c

    missing_cases = [i for i, x in enumerate(case_files_out) if x is None]
    if len(missing_cases) > 0: print('Warning  some cases file missing from dir ', dir_name, ', missing case files for case number(s) =', str( (np.asarray(missing_cases)).tolist()))

    # look for requested case
    if case is not None:
        if not ( 0 < case <= len(case_files_out) ):
            raise ValueError('Requested case number='+ str(case) + ' is outside range of available cases in dir' + dir_name + ', where there are ' + str(len(case_files_out)) + '  cases, (note first case is case = 1')
        if case_files_out[case-1] is None:
            raise ValueError('Requested case number='+ str(case) + ' is missing from dir = ' + dir_name + ', where there are ' + str(len(case_files_out)) + '  cases, (note first case is case =1)')
        case_files_out = case_files_out[case - 1]

    return case_files_out

def read_case_info_file(case_info_file_name):
    # load runInfo and given case case_infofiles into dict
    # runcase_info is used as input for all particle_plot methods
    # if dir of case info does not match that in the case_info.json, then it is changed to match given name

    if type(case_info_file_name) == list:
        print('Warning: case_info_file is a list from a parallel run, loading first case, ie case_info_file[0], if another case required use read_case_info_file(case_info_file[n]) ')
        case_info = json_util.read_JSON(case_info_file_name[0])
    else:
        case_info = json_util.read_JSON(case_info_file_name)

    # make case info output dir consistent with given file name
    case_info['output_files']['run_output_dir'] = path.dirname(case_info_file_name)
    case_info['output_files']['root_output_dir'] = path.dirname(case_info['output_files']['run_output_dir'])

    # force run out dir and root output dir to match that of case_info_file name if user has data in another folder name
    case_info['output_files']['run_output_dir']  = path.dirname(case_info_file_name)
    case_info['output_files']['root_output_dir'] = path.dirname(case_info['output_files']['run_output_dir'])
    return case_info

def load_track_data(case_info_file_name, var_list=None, release_group= None, fraction_to_read=None, track_file_number=1, run_output_dir=None, gridID=0):
    # load one track file from squeuence of what may be split files
    # todo load split track files into  dictionary

    case_info = read_case_info_file(case_info_file_name)

    track_file = path.join( case_info['output_files']['run_output_dir'], case_info['output_files']['tracks_writer'][track_file_number-1])
    tracks = read_ncdf_output_files.read_particle_tracks_file(track_file, var_list, release_group=release_group, fraction_to_read=fraction_to_read)
    tracks['grid'] = load_grid(case_info_file_name,gridID=gridID)

    tracks= _extract_useful_info(case_info, tracks)
    x= tracks['x'][:,:,0]
    y = tracks['x'][:, :, 1]
    tracks['axis_lim'] = np.asarray([np.nanmin(x), np.nanmax(x),np.nanmin(y), np.nanmax(y)])
    return tracks

def _extract_useful_info(case_info, d):
    # get release group info
    if 'version_info' in case_info and 'major' in case_info['version_info'] and case_info['version_info']['major'] >= 0.5:
        prg_info = read_ncdf_output_files.read_release_groups_info(path.join(case_info['output_files']['run_output_dir'] , case_info['output_files']['release_groups']))

    else:
        #todo deprecated from version 0.5
        prg_info = case_info['release_groups']

    d.update( particle_status_flags =case_info['particle_status_flags'],
              particle_release_groups= prg_info)

    return d

def load_concentration_data(case_info_file_name, name= None):
    case_info = read_case_info_file(case_info_file_name)
    nc_file_name= _get_role_dict_file_name(case_info, 'particle_concentrations', name)
    d = read_ncdf_output_files.read_concentration_file(nc_file_name)
    d['grid'] = load_grid(case_info_file_name)
    d =  _extract_useful_info(case_info, d)
    return d

def load_grid(case_info_file_name,gridID=0):
    # load OT output file grid from  output of load_runInfo() or load_runcase_info()
    case_info = read_case_info_file(case_info_file_name)
    if gridID==0:
        grid_file = path.join(case_info['output_files']['run_output_dir'], case_info['output_files']['grid'])
    else:
        grid_file = case_info['output_files']['nested_grids'][gridID-1]

    grid_file = path.join(case_info['output_files']['run_output_dir'],grid_file)
    d = read_ncdf_output_files.read_grid_file(grid_file)

    if 'grid_outline' not in d:
        # todo deprecated from version 0.5, data now in netcdf grid file
        # load  json grid outline and convert outline to numpy arrays
        grid_outline_file = path.join(case_info['output_files']['run_output_dir'], case_info['output_files']['grid_outline'])
        d['grid_outline'] = read_ncdf_output_files.read_grid_outline_file(grid_outline_file)

        # make outline list np arrays for plotting
        for key in d['grid_outline']['domain']:
            d['grid_outline']['domain'][key] = np.asarray(d['grid_outline']['domain'][key])
        for n in range(len(d['grid_outline']['islands'])):
            for key in  d['grid_outline']['islands'][n]:
                d['grid_outline']['islands'][n][key]= np.asarray( d['grid_outline']['islands'][n][key])

    return d

def load_stats_data(case_info_file_name, name = None,nt=None):
    # load gridded or polygon stas file using runcase_info, the output of  load_runcase_info()

    case_info = read_case_info_file(case_info_file_name)
    name = _get_role_dict_name(case_info, 'particle_statistics', name)
    stat_nc_file_name = _get_role_dict_file_name(case_info, 'particle_statistics', name)
    d = read_ncdf_output_files.read_stats_file(stat_nc_file_name,nt=nt)

    d = _extract_useful_info(case_info, d)
    d['grid'] = load_grid(case_info_file_name)
    return d

def load_residence_file(case_info_file_name=None,name=None, var_list=[]):
    # load residence time in relese polygon

    case_info = read_case_info_file(case_info_file_name)
    name = _get_role_dict_name(case_info, 'particle_statistics', name)
    nc_file_name =  _get_role_dict_file_name(case_info, 'particle_statistics', name)
    d = read_ncdf_output_files.read_residence_file(nc_file_name, var_list)
    #d['info']= case_info['class_roles_info']['particle_statistics'][name]
    d['params'] = case_info['working_params']['class_roles']['particle_statistics'][name]

    d = _extract_useful_info(case_info, d)
    d['grid'] = load_grid(case_info_file_name)
    return d

def load_LSC(case_info_file_name):
    # load LCS
    case_info = read_case_info_file(case_info_file_name)
    fn = path.join(case_info['output_files']['run_output_dir'],case_info['output_files']['integrated_model'])
    d = read_ncdf_output_files.read_LCS(fn)
    d = _extract_useful_info(case_info, d)
    d['grid'] = load_grid(case_info_file_name)
    return d

def _get_role_dict_name(caseinfo, class_dict, name= None):
    o = caseinfo['output_files']
    c = o[class_dict]
    if name is None:
        name = list(caseinfo['class_roles_info'][class_dict].keys())[0]  # use first one
        print('Post processing ,no name given loading "' + class_dict + '" named  "' + name + '"')
    if name not in c:
        raise Exception('Post processing error, "' + class_dict + '" does not have class name  "' + name + '"')
    return name

def _get_role_dict_file_name(caseinfo, class_dict, name=None):
    # val is astring name of relese group ot integer
    o = caseinfo['output_files']
    name =  _get_role_dict_name(caseinfo, class_dict, name= name)
    return path.join(o['run_output_dir'],o[class_dict][name])

# ubder dev
def dev_load_events_file(case_info_file_name, name=None):
    # load  flat events
    #todo finish
    case_info = read_case_info_file(case_info_file_name)
    nc_file_name = _get_role_dict_file_name(case_info, 'event_loggers', name)
